<?php
/**
 * Site Export Plugin
 *
 * @license    GPL 2 (http://www.gnu.org/licenses/gpl.html)
 * @author     i-net software <tools@inetsoftware.de>
 * @author     Gerry Weissbach <gweissbach@inetsoftware.de>
 */

// must be run within Dokuwiki
if(!defined('DOKU_INC')) define('DOKU_INC',realpath(dirname(__FILE__).'/../../../../').'/');
if(!defined('DOKU_PLUGIN')) {

    // Just for sanity
    require_once(DOKU_INC.'inc/plugin.php');
    define('DOKU_PLUGIN',DOKU_INC.'lib/plugins/');
}

if(!defined('_MPDF_TEMP_PATH')) define("_MPDF_TEMP_PATH", '/tmp/');
require_once(DOKU_PLUGIN.'action.php');
require_once(DOKU_PLUGIN.'siteexport/inc/httpproxy.php');
require_once(DOKU_INC.'/inc/search.php');
// require_once(DOKU_INC.'/inc/io.php');

class action_plugin_siteexport_ajax extends DokuWiki_Action_Plugin {

    var $zipFile = '';
    var $tempDir = '';
    var $fileType = 'html';
    var $fileChecked = array();
    var $depth = '';
    var $namespace = '';
    var $exportNamespace = '';
    var $mpdf = false;

    var $emptyNSToc = true;
    var $origEclipseZipFile = 'doc.zip';
    var $eclipseZipFile = '';
    var $addParams = false;
    var $origZipFile = '';
    var $downloadZipFile = '';
    var $pattern = '';
    var $exportLinkedPages = false;
    var $additionalParameters = array();
    var $isAuthed = false;
    var $debug = false;
    var $firstRE = true;
    var $isAJAX = false;

    /**
     * Debug Level
     * the level of what should be logged during the proxied session.
     * To activate the logging, you have to enter a loglevel below 5 (see below) to log
     * to the screen. If you use the debugFile option the logstream will be rerouted
     * to this file.
     *
     * Default: 5 / No logging
     *
     * Available DEBUG Levels:
     *  5 = off      - only socket exceptions will be shown to avoid blank pages
     *  4 = ERROR    - Log errors of the proxy process
     *  3 = WARN     - Log warnings during the proxy process
     *  2 = INFO     - Log information about the ongoing connection process
     *  1 = DEBUG    - detailed log about variable states
     *  0 = VERBOSE  - Additionally logs the reponse body from the server
     */
    var $debugLevel = 5;

    /**
     * Set a valid and writeable filename to have the debug information written into a file
     * Set the debugLevel below 5 to enable the debugging.
     *
     * e.g. $CC->debugFile = '/temp/ccproxy.txt';
     * e.g. $CC->debugFile = 'C:\temp\ccproxy.txt';
     */
    var $debugFile = '';

    function action_plugin_siteexport_ajax() {
        $this->debugLevel = $this->getConf("debugLevel");
        $this->debugFile = $this->getConf("debugFile");
         
        if ( empty($this->debugFile) ) {
            $this->debugFile = null;
        }
    }

    /**
     * for backward compatability
     * @see inc/DokuWiki_Plugin#getInfo()
     */
    function getInfo(){
        if ( method_exists(parent, 'getInfo')) {
            $info = parent::getInfo();
        }
        return is_array($info) ? $info : confToHash(dirname(__FILE__).'/../plugin.info.txt');
    }

    /**
     * Register Plugin in DW
     **/
    function register(&$controller) {
        $controller->register_hook('AJAX_CALL_UNKNOWN', 'BEFORE', $this, 'ajax_siteexport_provider');
        $controller->register_hook('ACTION_ACT_PREPROCESS', 'BEFORE', $this, 'siteexport_action');
        $controller->register_hook('MEDIA_SENDFILE', 'BEFORE', $this, 'siteexport_sendfile');

    }

    function __init_files($downloadZipFile=null) {
        global $ID;

        // Load Variables
        $this->origZipFile = $this->getConf('zipfilename');
        $this->downloadZipFile = $this->__getSpecialExportFileName($this->origZipFile, $downloadZipFile);
        $this->eclipseZipFile = $this->__getSpecialExportFileName(getNS($this->origZipFile) . ':' . $this->origEclipseZipFile, $this->pattern);
        $this->zipFile = mediaFN($this->downloadZipFile);
        $this->tempDir = mediaFN(getNS($this->origZipFile));
        $this->exportLinkedPages = intval($_REQUEST['exportLinkedPages']) == 1 ? true : false;

        $this->namespace = $this->getNamespaceFromID($_REQUEST['ns'], $PAGE);
        $this->addParams = !empty($_REQUEST['addParams']);

        if ( intval($_REQUEST['pdfExport']) == 1 && file_exists(DOKU_PLUGIN . 'dw2pdf/mpdf/mpdf.php') ) {
            require_once(DOKU_PLUGIN . 'siteexport/inc/mpdf.php');
            $this->mpdf = true;
            $this->fileType = 'pdf';
        }

        // set export Namespace - which is a virtual Root
        $pg = noNS($ID);
        if ( empty( $this->namespace ) ) { $this->namespace = $this->getNamespaceFromID(getNS($ID), $pg); }
        $this->exportNamespace = !empty($_REQUEST['ens']) && preg_match("%^" . $this->getNamespaceFromID($_REQUEST['ens'], $pg) . "%", $this->namespace) ? $this->getNamespaceFromID($_REQUEST['ens'], $pg) : $this->namespace;

        // Strip params that should be forwarded
        $this->additionalParameters = $_REQUEST;
        $this->__removeWikiVariables($this->additionalParameters, true);

        $tmpID = $ID;
        $ID = $this->origZipFile;

        $INFO = pageinfo();
        if ( $INFO['perm'] < AUTH_UPLOAD ) {
            list ( $USER, $PASS) = $this->__basic_authentication();
            auth_login($USER, $PASS);
        }

        $ID = $tmpID;
    }


    /**
     * Export via calling URL
     **/
    function siteexport_action( &$event ) {
        global $ID;

        if ( $event->data != 'siteexport' ) { return false; }
        if ( headers_sent() ) {
            msg("The siteexport function has to be called prior to any header output.", -1);
        }

        $this->_debug("========================================", null, 1);
        $this->_debug("Starting export from URL call", null, 1);

        $event->preventDefault();
        $event->stopPropagation();

        $this->__init_files(( !empty($_REQUEST['pattern']) && !empty($_REQUEST['startcounter']))? $_REQUEST['pattern'] : null);

        // Fake security Token if none given
        if ( empty( $_REQUEST['sectok'] ) ) {
            $_REQUEST['sectok'] = getSecurityToken();
        }

        $starttimer = time();
        $timerdiff = $this->getConf('max_execution_time');

        $data = $this->__get_siteexport_list_and_init_tocs($ID, !empty($_REQUEST['startcounter']));

        if ( $data === false ) {
            header("HTTP/1.0 401 Unauthorized");
            print 'Unauthorized';
            exit;
        }

        $counter = 0;

        if ( count($data) == 0 ) {
            exit();
        }

        foreach ( $data as $site ) {

            if ( empty($_REQUEST['startcounter'])  || $counter >= intval($_REQUEST['startcounter']) ) {
                $status = $this->__siteexport_add_site($site['id']);
            }

            $counter ++;
            if ( time() - $starttimer >= $timerdiff ) {
                $this->__startredirctprocess($counter);
            }

        }

        $this->_debug("Finishing export from URL call", null, 1);
        $this->_debug("========================================", null, 1);

        $this->cleanCacheFiles();
        //        exit(0);

        send_redirect(ml($this->origZipFile, array('cache' => 'nocache', 'siteexport' => $this->pattern, 'sectok' => getSecurityToken()), true, '&'));
    }

    /**
     * AJAX Provider
     **/
    function ajax_siteexport_provider(&$event, $args) {

        $this->isAJAX = true;

        switch( $event->data ) {
            case '__siteexport_getsitelist': $this->ajax_siteexport_getsitelist( $event ); break;
            case '__siteexport_addsite': $this->ajax_siteexport_addsite( $event ); break;
            case '__siteexport_generateurl': $this->ajax_siteexport_generateurl( $event ); break;
        }
    }

    /**
     * generate direct access URL
     **/
    function ajax_siteexport_generateurl( &$event ) {

        $event->preventDefault();
        $event->stopPropagation();

        $NS = $_REQUEST['ns'];
        $this->__removeWikiVariables($_REQUEST);
        $_REQUEST['do'] = 'siteexport';

        $additionalKeys = "";
        foreach( $_REQUEST as $key => $value ) {
            if ( !is_array($value) ) { continue; }
            foreach ( array_values($value) as $aValue ) {
                $additionalKeys .= "&$key" . "[]=$aValue";
            }

            unset($_REQUEST[$key]);
        }

        $url = wl($NS, $_REQUEST, true, '&') . $additionalKeys;

        // WGET Redirects - this is an option for wget only.
        // Calculate the maximum redirects that we want to allow. A Problem is that we don't know how long it will take to fetch one page
        // Therefore we assume it takes about 5s for each page - that gives the freedom to have anough time for redirect.
        $maxRedirectNumber = ceil( ( count($this->__get_siteexport_list($NS)) * 5) / $this->getConf('max_execution_time') );
        $maxRedirect = $maxRedirectNumber > 0 ? '--max-redirect=' . ($maxRedirectNumber+3) . ' ' : '';

        $this->_debug("Generating Direct Download URL", $url, 2);

        // If there was a Runtime Exception
        if ( !$this->firstRE ) {
            $this->_debug("There have been errors while generating the download URLs.", null, 4);
            return;
        }

        echo $url;
        echo "\n";
        echo 'wget ' . $maxRedirect . '--output-document=' . array_pop(explode(":", ($this->getConf('zipfilename')))) . ' --post-data="' . parse_url($url, PHP_URL_QUERY) . '" ' . wl($NS, null, true) . ' --http-user=USER --http-passwd=PASSWD';

        return;
    }

    /**
     * Get List of sites to be exported for AJAX (wrapper)
     **/
    function ajax_siteexport_getsitelist( &$event ) {

        $event->preventDefault();
        $event->stopPropagation();

        $this->__init_files();
        $data = $this->__get_siteexport_list_and_init_tocs($_REQUEST['ns']);

        // Important for reconaisance of the session

        if ( $data === false ) {
            $this->runtimeException("No data generated. List of Files is 'false'.");
            return;
        }

        if ( empty($data) ) {
            $this->runtimeException("Generated list is empty.");
            return;
        }

        // If there was a Runtime Exception
        if ( !$this->firstRE ) {
            $this->_debug("There have been errors while generating site list.", null, 4);
            return;
        }

        echo "$this->pattern\n";
        foreach($data as $line ){
            echo $line['id'] . "\n";
        }

        return;
    }

    /**
     * Add a page to the package (for AJAX calls - Wrapper)
     **/
    function ajax_siteexport_addsite( &$event ) {

        $event->preventDefault();
        $event->stopPropagation();

        $this->_debug("========================================", null, 1);
        $this->_debug("Starting export from AJAX call", null, 1);

        $this->__init_files($_REQUEST['pattern']);
        $status = $this->__siteexport_add_site($_REQUEST['site']);
        if ( $status === false ) { return; }

        $this->_debug("Finishing export from AJAX call", null, 1);
        $this->_debug("========================================", null, 1);

        // Print the download zip-File
        $this->cleanCacheFiles();

        // If there was a Runtime Exception
        if ( !$this->firstRE ) {
            $this->_debug("There have been errors during the export.", null, 4);
            return;
        }

        print ml($this->origZipFile, array('cache' => 'nocache', 'siteexport' => $this->pattern), true, '&');
        return;
    }

    /**
     * Fetch the list of pages to be exported
     **/
    function __get_siteexport_list($NS) {
        global $conf;

        $NS = $this->namespace = $this->getNamespaceFromID($NS, $PAGE);

        $depth = $this->getConf('depth');
        $query = '';
        $doSearch = 'search_allpages';

        switch( intval($_REQUEST['depthType']) ) {
            case 0:
                $query = $this->cleanID(str_replace(":", "/", $NS.':'.$PAGE));
                resolve_pageid($NS, $PAGE, $exists);

                if ( $exists ) {
                    return array( array( 'id' => $PAGE) );
                }
            case 1:	$depth = 0;
            break;
            case 2:	$depth = intval($_REQUEST['depth']);
            break;
        }

        $opts = array( 'depth' => $depth, 'skipacl' => $this->getConf('skipacl'), 'query' => $query);
        $data = array();
        require_once (DOKU_INC.'inc/search.php');

        // Check, which TOC to take
        if ( empty($_REQUEST['useTocFile']) ) {
            search($data, $conf['datadir'], $doSearch, $opts, $this->namespace);
        } else {
            $this->_debug("Using TOC for data", null, 2);

            $doSearch = 'search_pagename';

            // Create Data of the TOC File should be used instead
            $opts['query'] = 'toc.txt';
            search($data, $conf['datadir'], $doSearch, $opts, $this->namespace);
            $data = p_get_metadata($data[0]['id'], 'sitetoc siteexportTOC', true);
        }

        $this->_debug("Exporting the following sites: ", $data, 2);
        return $data;
    }

    function __get_siteexport_list_and_init_tocs($NS, $isRedirected=false ) {

        // Clean up if not redirected
        if ( !$isRedirected && !$this->__removeOldZip() ) {
            $this->runtimeException("Can't remove old files.");
            return false;
        }

        $data = $this->__get_siteexport_list($NS);

        // Create Eclipse Documentation Pages - TOC.xml, Context.xml
        if ( !empty($_REQUEST['absolutePath']) ) $this->namespace = "";
        $this->emptyNSToc = !empty($_REQUEST['emptyTocElem']);
        $this->__removeOldZip( $this->eclipseZipFile );

        if ( !empty($_REQUEST['eclipseDocZip']) ) {
            $this->_debug("Generating eclipseDocZip", null, 2);
            $this->__moveDataToZip($this->__getTOCXML($data), 'toc.xml');
            $this->__moveDataToZip($this->__getContextXML($data), 'context.xml');
        } else  if ( !empty($_REQUEST['JavaHelpDocZip']) ) {
            list($tocData, $mapData) = $this->__getJavaHelpTOCXML($data);
            $this->_debug("Generating JavaHelpDocZip", null, 2);
            $this->__moveDataToZip($tocData, 'toc.xml');
            $this->__moveDataToZip($mapData, 'map.xml');
        }

        return $data;
    }

    /**
     * Add page with ID to the package
     **/
    function __siteexport_add_site( $ID ) {
        global $conf, $currentID;

        // Which is the current ID?
        $currentID = $ID;

        $this->_debug("========================================", null, 2);
        $this->_debug("Adding Site: '$ID'", null, 2);

        if ( !empty($_REQUEST['eclipseDocZip']) || !empty($_REQUEST['JavaHelpDocZip']) ) {
            $this->zipFile = mediaFN( $this->eclipseZipFile );
        }

        // say, what to export and Build URL
        $do = (intval($_REQUEST['exportbody']) == 1 ? (empty($_REQUEST['renderer']) ? $conf['renderer_xhtml'] : $_REQUEST['renderer'] ) : '' );
        $do = ($do == $conf['renderer_xhtml'] && intval($_REQUEST['exportbody']) != 1) || ($do == 'pdf' && $this->mpdf) ? array() : array('do' => 'export_' . $do);

        if ( $do['do'] == 'export_' ) unset($do['do']);

        // set Template
        if ( !empty( $_REQUEST['template'] ) ) {
            $do['template'] = $_REQUEST['template'];
        }

        $ID = $this->cleanID($ID);
        $url = $this->wl($ID, $do, true, '&');

        // Parse URI PATH and add "html"
        $fileName = $this->__getSiteName($ID, true);

        $this->fileChecked[$url] = $fileName; // 2010-09-03 - One URL to one FileName
        $this->depth = str_repeat('../', count(explode('/', $fileName))-1);

        // fetch URL and save it in temp file
        $tmpFile = $this->__getHTTPFile($url);
        if ( $tmpFile === false ) { return $this->runtimeException("Creating temporary download file failed for '$url'. See log for more information."); }

        // If a Filename was given that does not comply to the original name, us this one!
        if ( !empty($tmpFile[1]) && !strstr($fileName, $tmpFile[1]) ) {

            $dParts = explode('/', $fileName);
            array_pop($dParts);
            $dParts[] = $tmpFile[1];

            $fileName = implode('/', $dParts);
            $this->fileChecked[$url] = $fileName;
        }

        // Add to zip
        $status = $this->__addFileToZip($tmpFile[0], $fileName);
        @unlink($tmpFile[0]);

        if ( !empty($_REQUEST['eclipseDocZip']) || !empty($_REQUEST['JavaHelpDocZip']) ) {
            $this->__addFileToZip($this->zipFile, $this->origEclipseZipFile, mediaFN($this->downloadZipFile));
        }

        return $status;
    }


    /**
     * Create name for the file inside the zip and the replacements
     **/
    function __shortenName($NAME) {
        global $conf;

        $NS = $this->exportNamespace;
        $NAME = preg_replace("%^" . DOKU_BASE . "%", "", $NAME);
        $NAME = preg_replace("%^(_media/)?(" . $NS . "/)?%", "", $NAME);

        $this->_debug("Shortening file to '$NAME'", null, 1);
        return $NAME;
    }

    /**
     * Download the file via HTTP URL + recurse if this is not an image
     * The file will be saved as temporary file. The filename is the result.
     **/
    function __getHTTPFile($URL, $RECURSE=false, $newAdditionalParameters=null) {
        global $conf;

        $EXCLUDE = str_replace('/', '\/', ($this->getConf('exclude')));

        if ( !empty($EXCLUDE) && preg_match("/($EXCLUDE)/i", $URL) ) { return false; }

        require_once( DOKU_INC . 'inc/HTTPClient.php');

        $http = new HTTPProxy($this);
        $http->max_bodysize = $conf['fetchsize'];

        // Add additional Params
        $this->__addAdditionalParametersToURL($URL, $newAdditionalParameters);

        $this->_debug("Fetching URL: '$URL'", null, 2);
        $getData = $http->get($URL);

        if( $getData === false ) {

            if ( $http->status < 200 || $http->status > 206 ) {
                $this->_debug("HTTP status was '{$http->status}'.", null , 4);
            } else {
                $this->_debug("Sending request failed.", $http->error, 4);
            }

            return false;
        }

        if( empty($getData) ) {
            $this->_debug("No data fetched.", null , 4);
            return false;
        }

        $tmpFile = tempnam($this->tempDir , 'siteexport__');

        $fp = fopen( $tmpFile, "w");
        if(!$fp) {
            $this->_debug("Can't open temporary File '$tmpFile'.", null , 4);
            return false;
        }

        if ( !$RECURSE ) {
            // Parse URI PATH and add "html"
            $this->_debug("========================================", null, 1);
            $this->_debug("Starting to recurse file '$URL'", null , 1);
            $this->_debug("========================================", null, 1);
            $this->__getInternalLinks($getData);
            $this->_debug("========================================", null, 1);
            $this->_debug("Finished to recurse file '$URL'", null , 1);
            $this->_debug("========================================", null, 1);
        }

        fwrite($fp,$getData);
        fclose($fp);

        return array($tmpFile, preg_replace("/.*?filename=\"?(.*?)\"?;?$/", "$1", $http->resp_headers['content-disposition']));
    }

    /**
     * Find internal links in the currently downloaded file. This also matches inside CSS files
     **/
    function __getInternalLinks(&$DATA) {

        $PATTERN = '(href|src|action)="([^"]*)"';
        $CALLBACK = array($this, '__fetchAndReplaceLink');
        $DATA = preg_replace_callback("/$PATTERN/i", $CALLBACK, $DATA);

        $PATTERNCSS = '(url\s*?)\(([^\)]*)\)';
        $DATA = preg_replace_callback("/$PATTERNCSS/i", $CALLBACK, $DATA);
    }

    /**
     * Deep Fetch and replace of links inside the texts matched by __getInternalLinks
     **/
    function __fetchAndReplaceLink($DATA) {
        global $conf, $currentID;

        $noDeepReplace = true;
        $newAdditionalParameters = $this->additionalParameters;
        $newDepth = $this->depth;
        $hadBase = false;

        $this->_debug("Starting Link Replacement", $DATA, 2);

        $DATA[2] = urldecode($DATA[2]);
        // External and mailto links
        if ( preg_match("%^(https?://|mailto:|javascript:)%", $DATA[2]) ) {
            $this->_debug("Don't like http, mailto or javascript links here", null, 1);
            return $this->__rebuildLink($DATA, "");
        }
        //if ( preg_match("%^(https?://|mailto:|" . DOKU_BASE . "/_export/)%", $DATA[2]) ) { return $this->__rebuildLink($DATA, ""); }
        // External media - this is deep down in the link, so we have to grep it out
        if ( preg_match("%media=(https?://.*?$)%", $DATA[2], $matches) ) {
            $DATA[2] = $matches[1];
            $this->_debug("This is an HTTP like somewhere else", $DATA, 1);
            return $this->__rebuildLink($DATA, "");
        }
        // reference only links won't have to be rewritten
        if ( preg_match("%^#.*?$%", $DATA[2]) ) {
            $this->_debug("This is a refercence only", null, 1);
            return $this->__rebuildLink($DATA, "");
        }

        // strip all things out
        // changed Data
        $PARAMS = @parse_url($DATA[2], PHP_URL_QUERY);
        $ANCHOR = @parse_url($DATA[2], PHP_URL_FRAGMENT);
        $DATA[2] = @parse_url($DATA[2], PHP_URL_PATH);

        // 2010-08-25 - fix problem with relative movement in links ( "test/../test2" )
        $tmpData2 = '';
        while( $tmpData2 != $DATA[2] ) {
            $tmpData2 = $DATA[2];
            $DATA[2] = preg_replace("#/(?!\.\.)[^\/]*?/\.\./#", '/', $DATA[2]);
        }

        $temp = preg_replace("%^" . DOKU_BASE . "%", "", $DATA[2]);
        if ( $temp != $DATA[2] ) {
            $DATA[2] = $temp;
            $hadBase = true; // 2010-08-23 Check if there has been a rewrite here that will have to be considered later on
        }

        $this->_debug("URL before rewriting option for others than 1", array($DATA, $PARAMS, $hadBase), 1);

        // Handle rewrites other than 1
        if ( !preg_match('$^/?lib/$', $DATA[2]) ) {
            if ( $conf['userewrite'] == 2 ) {
                $DATA[2] = $this->__getInternalRewriteURL($DATA[2]);
            } elseif ( $conf['userewrite'] == 0 ) {
                $this->__getParamsAndDataRewritten($DATA, $PARAMS);
            }
        }

        $this->_debug("URL before rewriting option", array($DATA, $PARAMS), 1);

        $ORIGDATA2 = $DATA;
        //        $ORIGDATA2 = $DATA[2]; // 08/10/2010 - this line required a $this->wl which may mess up with the base URL
        $this->_debug("OrigDATA is:", $ORIGDATA2, 1);

        // Generate ID
        $DATA[2] = str_replace('/', ':', $DATA[2]);

        // If Data was empty this must be the same file!;
        if ( empty( $DATA[2] ) ) {
            $DATA[2] = $currentID;
        }

        $ID = $DATA[2];
        $MEDIAMATCHER = "#(_media(/|:)|media=|_detail(/|:)|_export(/|:)|do=export_)#i"; // 2010-10-23 added "(/|:)" for the ID may not contain slashes anymore
        $ID = $this->cleanID($DATA[2], null, preg_match($MEDIAMATCHER, $DATA[2]) );
        //        $ID = $this->cleanID($DATA[2], null, strstr($DATA[2], 'media') ); // Export anpassung nun weiter unten

        //        $IDexists = page_exists($ID); // 08/10/2010 - Not needed. This will be done in the next block.
        //        $this->_debug("Current ID: '$ID' exists: '" . ($IDexists ? 'true' : 'false') . "' (will be set to 'false' anyway)", null, 1);

        $IDifIDnotExists = $ID; // 08/10/2010 - Save ID - with possible upper cases to preserve them
        $IDexists = false;

        $this->_debug("Resolving ID: '$ID'", null, 1);
        if ( preg_match($MEDIAMATCHER, $DATA[2]) ) {
            resolve_mediaid(null, $ID, $IDexists);
             
            $this->_debug("Current mediaID to filename: '" . mediaFN($ID) . "'", null, 1);
        } else {
            resolve_pageid(null, $ID, $IDexists);
            $this->_debug("Current ID to filename: '" . wikiFN($ID) . "'", null, 1);
        }

        $this->_debug("Current ID after resolvement: '$ID' the ID does exist: '" . ($IDexists ? 'true' : 'false') . "'", null, 1);
        //        $ORIGDATA2 = @parse_url($this->wl($ORIGDATA2, null, true)); // What was the next 2 line for? It did mess up with links from {{jdoc>}}
        //        $this->_debug("OrigData ID after parse:", $ORIGDATA2, 1); // 08/10/2010 - The lines are obsolete when the $ORIGDATA2 = $DATA. $ORIGDATA is only for fallback

        // 08/10/2010 - If the ID does not exist, we may have a problem here with upper cases - they will all be lower by now!
        if ( !$IDexists ) {
            $ID = $IDifIDnotExists; // there may have been presevered Upper cases. We will need them!
        }

        // $this->cleanID($DATA[2], null, strstr($DATA[2], 'media') || strstr($DATA[2], 'export') );
        if ( substr($ID, -1) == ':' || empty($ID) ) $ID .= $conf['start'];

        // Generate Download URL
        $PARAMS = trim(str_replace('&amp;', '&', $PARAMS));
        $this->__removeWikiVariables($PARAMS, false, true);

        $url = $this->wl($ID, null, true, null, null, true, $hadBase) . ( !empty( $ANCHOR) ? '#' . $ANCHOR : '' ) . ( !empty( $PARAMS) ? '?' . $PARAMS : '' );
        $this->_debug("URL from ID: '$url'", null, 1);

        // Parse URI PATH and add "html"
        $uri = @parse_url($url);
        $DATA[2] = $uri['path'];
        $DATA['ANCHOR'] = $ANCHOR;
        $DATA['PARAMS'] = $PARAMS;

        $this->_debug("DATA after parsing.", $DATA, 1);

        // Second Rewrite for UseRewrite = 2
        if ( $conf['userewrite'] == 2 ) {
            $DATA[2] = preg_replace( '$/lib/.*?fetch\.php$', '', $DATA[2]);
            $DATA[2] = preg_replace( '%(/lib/.*?detail\.php.*$)%', '\1' . '.' . $this->fileType, $DATA[2]);

            if ( preg_match( '%/(lib/.*?detail|doku)\.php%', $DATA[2])) {
                $noDeepReplace = false;
                $fileName = $this->__getSiteName($ID);
                $newDepth = str_repeat('../', count(explode('/', $fileName))-1);
            }

            $this->_debug("DATA after second rewrite with UseRewrite = 2", array($DATA, $noDeepReplace, $fileName, $newDepth), 1);
        }

        switch ( array_pop(explode('/', $DATA[2])) ) {
            // CSS Extra Handling with extra rewrites
            case 'css.php'	:	// $DATA[2] .=  ( !$this->addParams || empty($PARAMS) ? '' : '.' . $this->cleanID(preg_replace("/(=|\?|&amp;)/", ".", $PARAMS))) . '.css';
                $DATA[2] .=  '.' . $this->cleanID(preg_replace("/(=|\?|&amp;)/", ".", $PARAMS)) . '.css'; // allways put parameters behind
                // No paramters needed since they are rewritten.
                $DATA['PARAMS'] = "";
                $noDeepReplace = false;
                $fileName = $this->__getSiteName($ID);
                $newDepth = str_repeat('../', count(explode('/', $fileName))-1);
                $newAdditionalParameters['do'] = 'siteexport';

                $this->_debug("This is CSS file", array($DATA, $noDeepReplace, $fileName, $newDepth, $newAdditionalParameters), 2);

                break;
            case 'js.php'	:	// $DATA[2] .= ( !$this->addParams || empty($PARAMS) ? '' : '.' . $this->cleanID(preg_replace("/(=|\?|&amp;)/", ".", $PARAMS))) . '.js';
                $DATA[2] .=  '.t.' . $this->cleanID($_REQUEST['template']) . '.js'; // allways put parameters behind
                // set Template
                if ( !empty( $_REQUEST['template'] ) ) {
                    $url .= ( strstr($url, '?') ? '&' : '?' ) . 'template=' . $_REQUEST['template'];
                }
                // No paramters needed since they are rewritten.
                $DATA['PARAMS'] = "";
                $newAdditionalParameters['do'] = 'siteexport';

                $this->_debug("This is JS file", array($DATA, $url, $fileName, $newAdditionalParameters), 2);

                break;
                // Detail Handling with extra Rewrites if Paramaters are available - otherwise this is just the fetch
            case 'indexer.php' :
                $this->_debug("Skipping indexer", null, 2);
                return "";
                break;
            case 'detail.php' :
                $fileName = $this->__getSiteName($ID, true); // 2010-09-03 - rewrite with override enabled
            case 'doku.php' :
                if ( $this->addParams ) {
                    $noDeepReplace = false;

                    if ( empty($fileName) ) {
                        $fileName = $this->__getSiteName($ID); // 2010-09-03 - rewrite with override enabled
                    }

                    $newDepth = str_repeat('../', count(explode('/', $fileName))-1);
                    $this->__rebuildDataForNormalFiles($DATA, $PARAMS);

                    $this->_debug("This is doku.php or detail.php file with addParams", array($DATA, $fileName, $newDepth, $newAdditionalParameters), 2);
                    break;
                }

                $url = str_replace('detail.php', 'fetch.php', $url);
                $this->_debug("This is doku.php or detail.php file '$url'", null, 2);
                // Fetch Handling for media - rewriting everything
            case 'fetch.php':
                $this->__getParamsAndDataRewritten($DATA, $PARAMS, 'media');

                $DATA[2] = str_replace('/', ':', $DATA[2]);
                $ID = $this->cleanID($DATA[2], null, strstr($DATA[2], 'media'));

                $urlM = ml($ID, null, true);
                $uriM = @parse_url($urlM);
                $DATA[2] = $uriM['path'] . ( !empty( $ANCHOR) ? '#' . $ANCHOR : '' ) . ( !empty( $PARAMS) ? '?' . $PARAMS : '' );

                $DATA['PARAMS'] = "";
                $newAdditionalParameters = array();

                $this->_debug("This is fetch.php file", array($DATA, $ID), 2);
                break;

                // default Handling for Pages
            default			:
                if ( preg_match("%" . DOKU_BASE . "_detail/%", $DATA[2]) ) {

                    // GET ID Param from origdata2
                    preg_match("#id=(.*?)(&|\")#i", $DATA[0], $backlinkID);
                    $this->__rebuildDataForNormalFiles($DATA, $PARAMS);

                    $fileIDPart = isset($backlinkID[1]) && !empty($backlinkID[1]) ? $this->cleanID(urldecode($backlinkID[1])) : 'detail';

                    $DATA[2] .= '/' . $fileIDPart . '.' . $this->fileType; // add namespace and subpage for back button and add filetype

                    $noDeepReplace = false;
                    $fileName = $this->__shortenName($DATA[2]);
                    $newDepth = str_repeat('../', count(explode('/', $fileName))-1);
                    $url .= ( strstr($url, '?') ? '&' : '?' ) . 'id=' . $fileIDPart; // add id-part to URL for backlinks
                     
                    $DATA['PARAMS'] = "";

                    $this->_debug("This is something with '_detail' file", array($DATA, $backlinkID, $newDepth, $url), 2);
                } else if ( preg_match("%" . DOKU_BASE . "_export/(.*?)/%", $DATA[2], $fileType) ) {
                     
                    // Fixes multiple codeblocks in one file
                    $this->__rebuildDataForNormalFiles($DATA, $PARAMS);

                    // add the Params no matter what they are. This is export. We don't mess with other files
                    // adding the "/" fixes the usage of multiple codeblocks in the same namespace
                    $DATA[2] .= (empty( $PARAMS ) ? '' : '/' . $PARAMS) . '.'. $fileType[1];
                     
                    $DATA['PARAMS'] = "";
                    $this->_debug("This is something with '_export' file", $DATA, 2);

                } else if ( $IDexists ) { // 08/10/2010 - was page_exists($ID) - but this should do as well.
                    // If this is a page ... skip it!
                    $DATA[2] .= ( !$this->addParams || empty($PARAMS) ? '' : '.' . $this->cleanID(preg_replace("/(=|\?|&amp;)/", ".", $PARAMS)))  . '.' . $this->fileType;
                    $DATA[2] = $this->__shortenName($DATA[2]);

                    // If Parameters are to be included in the filename - they must not be added twice
                    if ( $this->addParams ) $DATA['PARAMS'] = "";

                    $this->_debug("This page really exists", $DATA, 3);
                    return $this->__rebuildLink($DATA);
                } else {
                    $this->__rebuildDataForNormalFiles($DATA, $PARAMS);
                }

                unset($newAdditionalParameters['diPlu']);
        }


        $this->_debug("DATA after SWITCH CASE decision", array($DATA, $noDeepReplace, $fileName, $newDepth), 1);

        if ( $this->mpdf ) {
            $this->__addAdditionalParametersToURL($url, $newAdditionalParameters);
            $DATA[2] = $url;
            unset($DATA['PARAMS']);
            $url = $this->__rebuildLink($DATA, '');

            $this->_debug("Creating PDF with URL '$url'", null, 2);

            return $url;
        }

        // Create Name to save the file at
        $DATA[2] = str_replace(':', '_', $DATA[2]);
        $DATA[2] = $this->__shortenName($DATA[2]);


        // File already loaded?
        // 2010-10-23 - changes in_array from DATA[2] to $url - to check real URLs, the DATA[2] file will be checked with fileExistsInZip
        if ( in_array($url, array_keys($this->fileChecked)) ) {
            $DATA[2] = $this->fileChecked[$url];
            $this->_debug("File has been checked before.", array($DATA, $url), 2);
            return $this->__rebuildLink($DATA);
        }

        // 2010-09-03 - second check if the file is in the ZIP already.
        if ( $this->__fileExistsInZip($DATA[2]) ) {
            $this->_debug("File with DATA exists in ZIP.", $DATA, 3);
            return $this->__rebuildLink($DATA);
        }

        // 2010-10-23 - What if this is a fetch.php? than we produced an error.
        //        $this->fileChecked[] = $DATA[2];
        $this->fileChecked[$url] = $DATA[2]; // 2010-09-03 - One URL to one FileName

        // get tempFile and save it
        $origDepth = $this->depth;
        $this->depth = $newDepth;

        $tmpID = $currentID;
        $tmpFile === false;

        $this->_debug("Going to get the file", array($url, $noDeepReplace, $newAdditionalParameters), 2);
        $tmpFile = $this->__getHTTPFile($url, $noDeepReplace, $newAdditionalParameters);
        $this->_debug("This is the getHTTPFile result", $tmpFile, 2);

        $currentID = $tmpID;
        $this->depth = $origDepth; // 2010-09-03 - Reset depth at the very end

        if ( $tmpFile === false ) {
            // Keep an potentially extra link intact

            $this->_debug("The fetched file '$url' is 'false'", null, 3);
            if ( $IDexists === false ) {
                $this->_debug("The file does not exist, fallback to ORIGDATA", $ORIGDATA2, 2);
                $DATA[2] = $this->__shortenName($ORIGDATA2[2]); // get Origdata Path
            }

            $this->fileChecked[$url] = $DATA[2]; // 2010-09-03 - One URL to one FileName
            $link = $this->__rebuildLink($DATA);
            $this->_debug("Final Link after empty file from '$url'", null, 2);

            return $link;
        }

        $this->_debug("The fetched file looks good.", $tmpFile, 1);

        // If a Filename was given that does not comply to the original name, us this one!
        if ( !empty($tmpFile[1]) && !strstr($DATA[2], $tmpFile[1]) ) {

            $dParts = explode('/', $DATA[2]);
            array_pop($dParts);
            $dParts[] = $tmpFile[1];

            $DATA[2] = implode('/', $dParts);
        }

        // Add to zip
        $this->fileChecked[$url] = $DATA[2]; // 2010-09-03 - One URL to one FileName

        $status = $this->__addFileToZip($tmpFile[0], $DATA[2]);
        @unlink($tmpFile[0]);

        $newURL = $this->__rebuildLink($DATA);
        $this->_debug("Returning final Link to document: '$newURL'", null, 2);

        return $newURL;
    }

    /**
     * Wrapper for fetching the Context or the TOC for Eclipse Documentation
     * This also puts the file into the zip package
     **/
    function __moveDataToZip($DATA, $FILENAME='toc.xml') {

        if ( empty($DATA) ) { return false; }

        $tmpFile = tempnam($this->tempDir , 'siteexport__');

        $fp = fopen( $tmpFile, "w");
        if(!$fp) return false;

        fwrite($fp,$DATA);
        fclose($fp);

        // Add to zip
        $status = $this->__addFileToZip($tmpFile, $FILENAME);
        @unlink($tmpFile);

        return true;
    }

    /**
     * Build the Eclipse Documentation TOC XML
     **/
    function __getJavaHelpTOCXML($DATA) {

        if ( count ( $DATA) == 0 ) {
            return false;
        }

        $TOCXML = "<?xml version=\"1.0\" encoding=\"utf-8\"?>\n<toc";
        $MAPXML = "<?xml version=\"1.0\" encoding=\"utf-8\"?>\n<map version=\"1.0\">";

        $startDepth = $DATA[0]['depth'];
        $currentLevel = 0;
        $check = array();

        // Go through the pages
        foreach ( $DATA as $elem ) {

            // support more than one view IDs ... for more than one reference
            $meta = p_get_metadata($elem['id'], 'context', true);
            if ( empty( $elem['name'] ) ) $elem['name'] = $this->__getSiteTitle($elem['id']);

            $elem['mapID'] = empty( $meta['id'] ) ? sectionID($elem['name'], $check) : $meta['id'];
            $elem['mapID'] = explode('|', $elem['mapID']);

            $elemLevel = $elem['depth'] - $startDepth;
            if ( $elemLevel <= $currentLevel && $currentLevel > 0 ) {
                $TOCXML .= "/";
            }

            $TOCXML .= ">";
            if ( $elemLevel < $currentLevel ) {
                $TOCXML .= "\n" . str_repeat("\t", $elemLevel+1) . "</tocitem>";
            }

            $currentLevel = $elemLevel;
            $TOCXML .= "\n" . str_repeat("\t", $elemLevel+1) . "<tocitem target=\"{$elem['mapID'][0]}\" text=\"" . $elem['name'] . "\"";

            foreach ( $elem['mapID'] as $VIEWID ) {
                $MAPXML .= "\n\t<mapID target=\"$VIEWID\" url=\"" . $this->__getSiteName($elem['id']) . "\"/>";
            }
        }

        $TOCXML .= "/>";
        for ( $currentLevel; $currentLevel > 0; $currentLevel --) {
            $TOCXML .= "\n" . str_repeat("\t", $currentLevel) . "</tocitem>";
        }

        $TOCXML .= "\n</toc>";
        $MAPXML .= "\n</map>";

        return array($TOCXML, $MAPXML);

    }

    /**
     * Build the Eclipse Documentation TOC XML
     **/
    function __getTOCXML($DATA) {

        $XML = "<?xml version=\"1.0\" encoding=\"utf-8\"?>\n<?NLS TYPE=\"org.eclipse.help.toc\"?>\n";

        $pagesArray = array();

        // Go through the pages
        foreach ( $DATA as $elem ) {

            $site = $elem['id'];
            $elems = explode('/', $this->__getSiteName($site));

            // Strip Site
            array_pop( $elems );
             
            // build the topic Tree
            $this->__buildTopicTree($pagesArray, $elems, $site);
        }

        $XML .= $this->__addXMLTopic($pagesArray, 'toc');

        return $XML;

    }

    /**
     * Load the topic Tree for the TOC - recursive
     **/
    function __buildTopicTree( &$PAGES, $DATA, $SITE ) {

        if ( empty( $DATA ) || !is_array($DATA) ) {
            // This is already a namespace
            if ( is_array($PAGES[noNS($SITE)]) ) {
                // The root already exists!
                if ( !empty($PAGES[noNS($SITE)][noNS($SITE)]) ) {
                    if ( strstr($PAGES[noNS($SITE)][noNS($SITE)], $SITE) ) {
                        // The SITE is in the parent Namespace, and the current Namespace has an index with same name
                        $PAGES['__' . noNS($SITE)] = $SITE;
                    } else {
                        $PAGES['__' . noNS($SITE)] = $PAGES[noNS($SITE)][noNS($SITE)];
                        $PAGES[noNS($SITE)][noNS($SITE)] = $SITE;
                    }
                } else {
                    $PAGES[noNS($SITE)][noNS($SITE)] = $SITE;
                }
            } else {
                // just a Page
                $PAGES[noNS($SITE)] = $SITE;
            }
            return;
        }

        $NS = array_shift($DATA);
        if ( !is_array( $PAGES[$NS] ) ) $PAGES[$NS] = empty($PAGES[$NS]) ? array() : array($PAGES[$NS]);
        $this->__buildTopicTree( $PAGES[$NS], $DATA, $SITE );

        return;
    }

    /**
     * Build the Topic Tree for TOC.xml
     **/
    function __addXMLTopic($DATA, $ITEM='topic', $LEVEL=0, $NODENAME='') {
        global $conf;

        $DEPTH = str_repeat("\t", $LEVEL);

        if ( !is_array($DATA) ) {
            return $DEPTH . '<' . $ITEM . ' label="' . $this->__getSiteTitle($DATA) . '" ' . ($ITEM != 'topic' ? 'topic' : 'href' ) . '="' . $this->__getSiteName($DATA) . "\" />\n";
        }
        // Is array from this point on
        list($indexTitle, $indexFile) = $this->__getIndexItem($DATA, $NODENAME);

        if ( empty( $indexTitle) ) $indexTitle = $this->__getSiteTitle( $conf['start'] );
        if ( !empty( $indexFile) ) $indexFile = ($ITEM != 'topic' ? 'topic' : 'href' ) . "=\"$indexFile\"";

        $isEmptyNode = count($DATA) == 1 && empty($indexFile);

        if ( !$isEmptyNode && ($this->emptyNSToc || count($DATA) > 0) )
        $XML = "$DEPTH<$ITEM label=\"$indexTitle\" $indexFile>";

        if ( !$isEmptyNode && count ($DATA) > 0 ) $XML .= "\n";

        foreach( $DATA as $NODENAME => $NS ) {

            $XML .= $this->__addXMLTopic($NS, ( !($this->emptyNSToc || count($DATA) > 1) && $ITEM != 'topic' ? $ITEM : 'topic' ), $LEVEL+(!$isEmptyNode ? 1 : 0), $NODENAME);

        }

        if ( !$isEmptyNode && count ($DATA) > 0 ) $XML .= "$DEPTH";
        if ( !$isEmptyNode && ($this->emptyNSToc || count($DATA) > 0) ) {
            $XML .= "</$ITEM>\n";
        }

        return $XML;
    }


    /**
     * Get the context XML
     **/
    function __getContextXML($DATA) {

        $XML = "<?xml version=\"1.0\" encoding=\"utf-8\"?>\n<?NLS TYPE=\"org.eclipse.help.context\"?>\n<contexts>\n";

        foreach ( $DATA as $elem ) {
            $ID = $elem['id'];
            $meta = p_get_metadata($ID, 'context', true);
            if ( empty( $meta['id'] ) ) { continue; }

            $TITLE = empty( $meta['title'] ) ? $this->__getSiteTitle($ID) : $meta['title'];

            // support more than one view IDs ... for more than one reference
            $VIEWIDs = empty( $meta['id'] ) ? $this->_xmlEntities(array_pop(explode(':', $ID))) : $meta['id'];
            $VIEWIDs = explode('|', $VIEWIDs);

            $DESCRIPTION = $this->_xmlEntities(p_get_metadata($ID, 'description abstract'));

            // Build topic Links
            $TOPICS = array( $this->__getSiteName($ID) => $TITLE . " (Details)" );
            $REFS = p_get_metadata($ID, 'relation references', true);
            if ( is_array($REFS) )
            foreach ( $REFS as $REL => $EXISTS ) {
                if ( !$EXISTS ) { continue; }
                $TOPICS[$this->__getSiteName($REL)] = $this->__getSiteTitle($REL);
            }

            // build XML - include multi view IDs
            foreach ( $VIEWIDs as $VIEWID ) {
                $XML .= "\t<context id=\"$VIEWID\" title=\"$TITLE\">\n";
                $XML .= "\t\t<description>$DESCRIPTION</description>\n";

                foreach ( $TOPICS as $URL => $LABEL ) {
                    $XML .= "\t\t<topic label=\"$LABEL\" href=\"$URL\" />\n";
                }

                $XML .= "\t</context>\n";
            }
        }

        $XML .= "</contexts>";
        return $XML;

    }

    /**
     * Determine if this is an index - and if so, find its Title
     **/
    function __getIndexItem(&$DATA, $NODENAME='') {
        global $conf;

        if ( !is_array($DATA) ) { return; }

        $indexTitle = '';
        $indexFile = '';
        foreach ( $DATA as $NODE => $indexSearch) {
            // Skip next Namespaces
            if ( is_array($indexSearch) ) { continue; }

            // Skip if this is not a start
            if ( $NODE != $conf['start'] ) { continue; }

            $indexTitle = $this->__getSiteTitle( $indexSearch );
            $indexFile = $indexSearch;
            unset($DATA[$NODE]);
            break;
        }

        if ( empty($indexFile) && !empty($DATA[$NODENAME]) ) {
            $indexTitle = $this->__getSiteTitle( $DATA[$NODENAME] );
            $indexFile = $DATA[$NODENAME];
            unset($DATA[$NODENAME]);
        }

        return array($indexTitle, $this->__getSiteName($indexFile));
    }


    /**
     * create a file name for the page
     **/
    function __getSiteName($ID, $overrideRewrite=false) {
        global $conf;

        if ( empty($ID) ) return false;

        $url = $this->wl($this->cleanID($ID), null, true, null, null, $overrideRewrite); // this must be done with rewriting set to override
        //$url = $this->wl($this->cleanID($ID), null, true); // this must be done with rewriting set to override
        $uri = @parse_url($url);
        if ( $uri['path'][0] == '/' ) {
            $uri['path'] = substr($uri['path'], 1);
        }

        return $this->__shortenName($uri['path'] . '.' . $this->fileType);
    }

    /**
     * get the Title for the page
     **/
    function __getSiteTitle($ID) {
        if (useHeading('content') && $ID) {
            $heading = p_get_first_heading($ID,true);
            if ($heading) {
                return $this->_xmlEntities($heading);
            }
        }
        return $this->_xmlEntities(array_pop(explode(':', $ID)));
    }

    /**
     * Encoding ()taken from DW - but without needing the renderer
     **/
    function _xmlEntities($string) {
        return htmlspecialchars($string,ENT_QUOTES,'UTF-8');
    }


    /**
     * build the new link to be put in place for the donwloaded site
     **/
    function __rebuildLink($DATA, $DEPTH = null) {

        // depth is set, skip this one
        if ( is_null( $DEPTH ) ) $DEPTH = $this->depth;
        $DATA[2] .= ( !empty( $DATA['PARAMS']) ? '?' . $DATA['PARAMS'] : '' ) . ( !empty( $DATA['ANCHOR'] ) ? '#' . $DATA['ANCHOR'] : '' );

        $newURL = $DATA[1] == 'url' ? $DATA[1] . '(' . $DEPTH . $DATA[2] . ')' : $DATA[1] . '="' . $DEPTH . $DATA[2] . '"';
        $this->_debug("Re-created URL: '$newURL'", null, 1);

        return $newURL;
    }

    /**
     * add a file to a zip file
     **/
    function __addFileToZip($FILE, $NAME, $ZIP=null) {

        if ( $NAME[0] === "/" ) {
            $this->_debug("Weird, the NAME for the ZIP started with a '/'. This may result in wrong links!", null, 3);
            $NAME = substr($NAME, 1);
        }

        // check for mpdf
        if ( $this->mpdf ) {
            $this->_debug("Trying to create PDF from File '$FILE' with name '$NAME'", null, 2);
            
            if ( $this->debugLevel <= 1 ) { // 2011-01-12 Write HTML to ZIP for Debug purpose
                $this->__writeFileToZip($FILE, "_debug/$NAME.html", $ZIP);
            }
            
            if ( !$this->createPDFFromFile($FILE, $NAME) ) {
                $this->_debug("Create PDF from File '$FILE' with name '$NAME' went wrong and is not being added!", null, 4);
                return;
            }
        }

        return $this->__writeFileToZip($FILE, $NAME, $ZIP);
    }
    
    function __writeFileToZip($FILE, $NAME, $ZIP) {
        if ( empty( $ZIP ) ) $ZIP = $this->zipFile;

        if ( !class_exists('ZipArchive') ) {
            $this->runtimeException("PHP class 'ZipArchive' does not exist. Please make sure that you have the ziplib extension for PHP installed.");
            return false;
        }

        $zip = new ZipArchive;
        if ( !$zip ) {
            $this->runtimeException("Can't create new instance of 'ZipArchive'. Please make sure that you have the ziplib extension for PHP installed.");
            return false;
        }

        $code = $zip->open($ZIP, ZipArchive::CREATE);
        if ($code === TRUE) {

            $this->_debug("Adding file '$NAME' to ZIP {$this->zipFile}", null, 2);

            $zip->addFile($FILE, $NAME);
            $zip->close();

            // ALibi Touching
            io_saveFile(mediaFN($this->origZipFile), "alibi file");

            return true;
        }

        $this->runtimeException("Zip Error #$code");
        return false;
    }

    /**
     * check if a file exists allready
     **/
    function __fileExistsInZip($NAME) {

        if ( empty( $ZIP ) ) $ZIP = $this->zipFile;
        $zip = new ZipArchive;
        $code = $zip->open($ZIP, ZipArchive::CREATE);
        if ($code === TRUE) {
            return !($zip->statName($NAME) === FALSE);
        }

        return false;
    }

    /**
     * remove an old zip file
     **/
    function __removeOldZip( $FILENAMEID=null, $checkForMore=true ) {
        global $INFO;
        global $conf;

        $returnValue = true;

        if ( empty($FILENAMEID) ) $FILENAMEID = $this->origZipFile;

        if ( !file_exists(mediaFN($FILENAMEID)) ) {
            $returnValue = true;
        } else {
            $INFO = pageinfo();
            if ( $INFO['perm'] < AUTH_DELETE && !$this->isAuthed ) {
                list ( $USER, $PASS) = $this->__basic_authentication();
                auth_login($USER, $PASS);
                $this->isAuthed = true;
            }

            require_once( DOKU_INC . 'inc/media.php');
            if ( !media_delete($FILENAMEID, $INFO['perm']) ) {
                $returnValue = false;
            }
        }

        if ( $checkForMore ) {
            // Try to remove more files.
            $ns = getNS($FILENAMEID);
            $fn = $this->__getSpecialExportFileName(noNS($FILENAMEID), '.+');

            $data = array();
            search($data, $conf['mediadir'], 'search_media', array('pattern' => "/$fn$/i"), $ns);

            if ( count($data > 0) ) {

                // 30 Minuten Cache Zeit
                $cache = (60*30);
                foreach ( $data as $media ) {

                    //decide if has to be deleted needed:
                    if( $media['mtime'] < time()-$cache) {
                        $this->__removeOldZip($media['id'], false);
                    }
                }
            }

        }

        return $returnValue;
    }

    function __getSpecialExportFileName($FILE, $PATTERN=null) {

        if ( empty($PATTERN) ){
            $PATTERN = md5(microtime(false));
        }

        // Set Pattern Global for other stuff
        if ( empty($this->pattern) ) $this->pattern = $PATTERN;

        $FA = explode('.', $FILE);
        $EXT = array_pop($FA);
        array_push($FA, 'auto');
        array_push($FA, $PATTERN);
        array_push($FA, $EXT);

        return implode('.', $FA);
    }

    /**
     * authenticate for direct downloads
     **/
    function __basic_authentication() {
        if (!isset($_SERVER['PHP_AUTH_USER'])) {
            header('WWW-Authenticate: Basic realm="Siteexport Authentication"');
            header('HTTP/1.0 401 Unauthorized');
            print 'Unauthorized'; // print has to stay here
            exit;
        }

        return array($_SERVER['PHP_AUTH_USER'], $_SERVER['PHP_AUTH_PW']);
    }

    /**
     * if confrewrite is set to internal rewrite, use this function - taken from a DW renderer
     **/
    function __getInternalRewriteURL($url) {
        global $conf;

        //construct page id from request URI
        if( $conf['userewrite'] != 2) { return $url; }

        //get the script URL
        if($conf['basedir']) {
            $relpath = '';
            $script = $conf['basedir'].$relpath.basename($_SERVER['SCRIPT_FILENAME']);
        } elseif($_SERVER['DOCUMENT_ROOT'] && $_SERVER['SCRIPT_FILENAME']){
            $script = preg_replace ('/^'.preg_quote($_SERVER['DOCUMENT_ROOT'],'/').'/','',
            $_SERVER['SCRIPT_FILENAME']);
            $script = '/'.$script;
        }else{
            $script = $_SERVER['SCRIPT_NAME'];
        }

        //clean script and request (fixes a windows problem)
        $script  = preg_replace('/\/\/+/','/',$script);
        $request = preg_replace('/\/\/+/','/',$url);

        //remove script URL and Querystring to gain the id
        if(preg_match('/^'.preg_quote($script,'/').'(.*)/',$request, $match)){
            $id = preg_replace ('/\?.*/','',$match[1]);
        }
        $id = urldecode($id);
        //strip leading slashes
        $id = preg_replace('!^/+!','',$id);

        return $id;
    }

    /**
     * rewrite parameter calls
     **/
    function __getParamsAndDataRewritten(&$DATA, &$PARAMS, $IDKEY='id') {

        $PARRAY = explode('&', str_replace('&amp;', '&', $PARAMS) );
        $PARAMS = "";

        foreach ( $PARRAY as $item ) {
            list($key, $value) = explode('=', $item, 2);
            if ( empty($key) || empty($value) )
            continue;

            if ( strtolower(trim($key)) == $IDKEY ) {
                $DATA[2] = preg_replace("%^" . DOKU_BASE . "%", "", $value);
                continue;
            }

            if ( !empty( $PARAMS) ) {
                $PARAMS .= '&';
            }

            $PARAMS .= "$key=$value";
        }
    }

    /**
     * rewrite detail.php calls
     **/
    function __rebuildDataForNormalFiles(&$DATA, &$PARAMS) {
        $PARTS = explode('.', $DATA[2]);
        if ( count($PARTS) > 1 ) {
            $EXT = '.' . array_pop($PARTS);
        }

        $PARAMS = preg_replace("/(=|\?|&amp;)/", ".", $PARAMS);
        $DATA[2] = implode('.', $PARTS) . ( !$this->addParams || empty($PARAMS) ? '' : '.' . $this->cleanID($PARAMS)) . ( $EXT == '.php' ? '.' . $this->fileType : $EXT );
        $DATA[2] = preg_replace("/\.+/", ".", $DATA[2]);
    }

    /**
     * This builds a link to a wikipage - changed for internal use here
     *
     * It handles URL rewriting and adds additional parameter if
     * given in $more
     *
     * @author Andreas Gohr <andi@splitbrain.org>
     */

    function wl($id='',$more='',$abs=false,$sep='&amp;', $IDexists=true, $overrideRewrite=false, $hadBase=false){
        global $conf;

        $this->_debug("Starting to build WL-URL for '$id'", null, 1);

        if(is_array($more)){
            $more = buildURLparams($more,$sep);
        }else{
            $more = str_replace(',',$sep,$more);
        }

        $id = idfilter($id);

        if($abs){
            $xlink = DOKU_URL;
            if ( !$IDexists && !$hadBase ) { // If the file does not exist, we have to remove the base. This link my be one to an parallel BASE.
                $xlink = preg_replace('#' . DOKU_BASE . '$#', '', $xlink);
            }
        }else if ($IDexists || $hadBase) { // if the ID does exist, we may add the base.
            $xlink = DOKU_BASE;
        } else{
            $xlink = "";
        }

        $xlink = preg_replace('#(?<!http:|https:)//+#','/', ($abs ? '' : '/') . "$xlink/"); // ensure slashes at beginning and ending, but strip doubles


        if ( $overrideRewrite ) {
            $this->_debug("Override enabled.", null, 1);
            $id = strtr($id,':','/');

            $xlink .= $id;
            if($more) $xlink .= '?'.$more;
        } else {
            if($conf['userewrite'] == 2 ){
                $xlink .= DOKU_SCRIPT.'/'.$id;
                if($more) $xlink .= '?'.$more;
            }elseif($conf['userewrite'] ){
                $xlink .= $id;
                if($more) $xlink .= '?'.$more;
            }elseif($id){
                $xlink .= DOKU_SCRIPT.'?id='.$id;
                if($more) $xlink .= $sep.$more;
            }else{
                $xlink .= DOKU_SCRIPT;
                if($more) $xlink .= '?'.$more;
            }
        }

        $this->_debug("internal WL function result: '$xlink'", null, 2);

        return $xlink;
    }

    /**
     * Remove unwanted chars from ID
     *
     * Cleans a given ID to only use allowed characters. Accented characters are
     * converted to unaccented ones
     *
     * @author Andreas Gohr <andi@splitbrain.org>
     * @param  string  $raw_id    The pageid to clean
     * @param  boolean $ascii     Force ASCII
     * @param  boolean $media     Allow leading or trailing _ for media files
     */
    function cleanID($raw_id,$ascii=false,$media=false){
        global $conf;
        global $lang;
        static $sepcharpat = null;

        global $cache_cleanid;
        $cache = & $cache_cleanid;

        // check if it's already in the memory cache
        if (isset($cache[(string)$raw_id])) {
            return $cache[(string)$raw_id];
        }

        $sepchar = $conf['sepchar'];
        if($sepcharpat == null) // build string only once to save clock cycles
        $sepcharpat = '#\\'.$sepchar.'+#';

        $id = trim((string)$raw_id);
        //        $id = utf8_strtolower($id); // NO LowerCase for us!

        //alternative namespace seperator
        $id = strtr($id,';',':');
        if($conf['useslash']){
            $id = strtr($id,'/',':');
        }else{
            $id = strtr($id,'/',$sepchar);
        }

        if($conf['deaccent'] == 2 || $ascii) $id = utf8_romanize($id);
        if($conf['deaccent'] || $ascii) $id = utf8_deaccent($id,-1);

        //remove specials
        $id = utf8_stripspecials($id,$sepchar,'\*');

        if($ascii) $id = utf8_strip($id);

        //clean up
        $id = preg_replace($sepcharpat,$sepchar,$id);
        $id = preg_replace('#:+#',':',$id);
        $id = ($media ? trim($id,':.-') : trim($id,':._-'));
        $id = preg_replace('#:[:\._\-]+#',':',$id);

        $cache[(string)$raw_id] = $id;
        return($id);
    }

    function __addAdditionalParametersToURL(&$URL, $newAdditionalParameters=null) {
         
        // Add additionalParameters
        if ( is_null($newAdditionalParameters) ) $newAdditionalParameters = $this->additionalParameters;
        if ( !empty($newAdditionalParameters) ) {
            foreach($newAdditionalParameters as $key => $value ) {
                if ( empty($key) || empty($value) ) { continue; }

                $append = '';

                if ( is_array($value) ) {
                    foreach( array_values($value) as $aValue ) { // Array Handling
                        $URL .= (strstr($URL, '?') ? '&' : '?') . $key . "[]=$aValue";
                    }
                } else {
                    $append = "$key=$value";
                    $URL .= empty($append) || strstr($URL, $append)  ? '' : (strstr($URL, '?') ? '&' : '?') . $append;
                }

            }
        }
    }

    function __startredirctprocess($counter) {
        global $ID;

        $URL = wl($ID);

        $this->additionalParameters = $_REQUEST;
        $this->additionalParameters['startcounter'] = $counter;
        $this->additionalParameters['pattern'] = $this->pattern;

        unset($this->additionalParameters['id']);
        unset($this->additionalParameters['u']);
        unset($this->additionalParameters['p']);
        unset($this->additionalParameters['r']);
        unset($this->additionalParameters['http_credentials']);

        $this->__addAdditionalParametersToURL($URL);

        send_redirect($URL);
        exit(0);
    }

    // Remove certain fields from the list.
    function __removeWikiVariables(&$removeArray, $advanced=false, $isString=false) {
        global $plugin_controller;

        if ( $isString && !is_array($removeArray) ) {
            $intermediate = str_replace("&amp;", "&", $removeArray);

            $removeArray = array();
            foreach( explode("&", $intermediate) as $param ) {
                list($key, $value) = explode("=", $param, 2);
                if ( empty($key) ) { continue; } // Don't check on Value, because there may be only the key that should be preserved

                if ( substr($key, -2) == '[]' ) {
                    $removeArray[substr($key, 0, -2)] = $value; // Array Handling
                } else {
                    $removeArray[$key] = $value;
                }
            }
        }

        if ( !empty($removeArray['diPlu']) ) {

            $allPlugins = array();
            foreach($plugin_controller->getList(null,true) as $plugin ) {
                // check for CSS or JS
                if ( !file_exists(DOKU_PLUGIN."$plugin/script.js") && !file_exists(DOKU_PLUGIN."$p/style.css") ) { continue; }
                $allPlugins[] = $plugin;
            }

            if ( count($removeArray['diPlu']) > (count($allPlugins) / 2) ) {
                $removeArray['diInv'] = 1;
                $removeArray['diPlu'] = array_diff($allPlugins, $removeArray['diPlu']);
            }
        }

        // 2010-08-23 - If there is still the media set, retain the id for e.g. detail.php
        if ( !isset($removeArray['media']) ) {
            unset($removeArray['id']);
        }

        unset($removeArray['do']);
        unset($removeArray['ns']);
        unset($removeArray['call']);
        unset($removeArray['sectok']);
        unset($removeArray['rndval']);
        unset($removeArray['tseed']);
        unset($removeArray['http_credentials']);
        unset($removeArray['u']);
        unset($removeArray['p']);
        unset($removeArray['r']);
        unset($removeArray['base']);
        unset($removeArray['siteexport']);
        unset($removeArray['DokuWiki']);

        if ( $removeArray['renderer'] == 'xhtml' ) {
            $removeArray['do'] = 'export_' . $removeArray['renderer'];
            unset($removeArray['renderer']);
        }


        if ( $advanced ) {
            if ( $removeArray['renderer'] != 'xhtml' && !empty($removeArray['renderer']) ) {
                $removeArray['do'] = 'export_' . $removeArray['renderer'];
            }


            // 2010-08-25 - Need fakeMedia for some _detail cases with rewrite = 2
            if ( isset($removeArray['fakeMedia'])  ) {
                unset($removeArray['media']);
                unset($removeArray['fakeMedia']);
            }

            unset($removeArray['ens']);
            unset($removeArray['renderer']);
            unset($removeArray['site']);
            unset($removeArray['namespace']);
            unset($removeArray['exportbody']);
            unset($removeArray['addParams']);
            unset($removeArray['template']);
            unset($removeArray['pattern']);
            unset($removeArray['eclipseDocZip']);
            unset($removeArray['useTocFile']);
            unset($removeArray['JavaHelpDocZip']);
            unset($removeArray['depth']);
            unset($removeArray['depthType']);
            unset($removeArray['startcounter']);
            unset($removeArray['pattern']);

        }

        if ( $isString && is_array($removeArray) ) {
            $intermediate = $removeArray;
            $removeArray = array();

            foreach ( $intermediate as $key => $value ) {
                if ( is_array($value) ) {
                    foreach( array_values($value) as $aValue ) { // Array Handling
                        $removeArray[] = $key . "[]=$aValue";
                    }
                } else {
                    $value = trim($value);

                    $removeArray[] = "$key" . ( ((empty($value) && intval($value) !== 0)) || $value == '' ? '' : "=$value" ); // If the Value is empty, the Key must be preserved
                }
            }

            $removeArray = implode("&", $removeArray);
        }
    }

    function getNamespaceFromID($NS, &$PAGE) {
        global $conf;
        // Check current page - if its an NS add the startpage
        $clean = true;
        resolve_pageid(getNS($NS), $NS, $clean);
        if ( ! page_exists($NS) && array_pop(explode(':', $NS)) != strtolower($conf['start'] )) { // Compare to lowercase since clean lowers it.
            $NS .= ':' . $conf['start'];
            resolve_pageid(getNS($NS), $NS, $clean);
        }

        $PAGE = noNS($NS);
        $NS = getNS($NS);

        return utf8_encodeFN(str_replace(':', '/', $NS));
    }

    /*
     * Clean JS and CSS cache files
     */
    function cleanCacheFiles() {

        $_SERVER['HTTP_HOST'] = preg_replace("/:?\d+$/", '', $_SERVER['HTTP_HOST']);
        $cache = getCacheName('scripts'.$_SERVER['HTTP_HOST'].'-siteexport-js-'.$_SERVER['SERVER_PORT'],'.js');
        $this->unlinkIfExists($cache);

        $tpl = trim(preg_replace('/[^\w-]+/','',$_REQUEST['template']));
        if($tpl){
            $tplinc = DOKU_INC.'lib/tpl/'.$tpl.'/';
            $tpldir = DOKU_BASE.'lib/tpl/'.$tpl.'/';
        }else{
            $tplinc = DOKU_TPLINC;
            $tpldir = DOKU_TPL;
        }

        // The generated script depends on some dynamic options
        $cache = getCacheName('styles'.$_SERVER['HTTP_HOST'].'-siteexport-js-'.$_SERVER['SERVER_PORT'].DOKU_BASE.$tplinc.$style,'.css');
        $this->unlinkIfExists($cache);
    }

    function unlinkIfExists($cache) {
        if ( file_exists($cache) ) {
            @unlink($cache);
            if(function_exists('gzopen')) @unlink("$cache.gz");
        }
    }

    /*
     * Redirect File to real File
     */
    function siteexport_sendfile(&$event, $args) {
        global $MEDIA;

        if ( empty($_REQUEST['siteexport']) ) {
            return;
        }

        $event->data['file'] = mediaFN($this->__getSpecialExportFileName($MEDIA, $_REQUEST['siteexport']));
        return;
    }

    function createPDFFromFile($filename, &$NAME) {

        if ( !preg_match("/" . $this->fileType . "$/", $NAME) ) {
            $this->_debug("Filetype {$this->fileType} did not match filename '$NAME'", null, 4);
            return false;
        }
        
        $mpdf = new siteexportPDF('UTF-8-s', $this);
        
        if ( !$mpdf ) {
            $this->_debug("Could not instantiate MPDF", null, 4);
            return false;
        }
        

        $html = file_get_contents($filename);

        if ( !strstr($html, "<html") ) {
            $this->_debug("Filecontent had no HTML starting tag", null, 4);
            return false;
        }

        // Save HTML too
        $this->_debug("Arranging HTML", null, 2);
        $this->arrangeHtml($html, 'bl,acronym');
        $this->_debug("Done arranging HTML:", $html, 1);

        $fp = fopen($filename, "w");
        fwrite($fp, $html);
        fclose($fp);
/*
        $fp = fopen("/tmp/siteexport.pdf.html", "w");
        fwrite($fp, $html);
        fclose($fp);
*/
        /*
         $this->mpdf = null;

         // change Name for extra file
         $fn = explode('.',$NAME);
         array_pop($fn);
         $fn[] = 'html';

         $this->__addFileToZip($filename, implode('.', $fn));
         $this->mpdf = $mpdf;
         /**/
        $mpdf->debug = true;
        $mpdf->list_indent_first_level = 1; // Indents the first level of lists.
        $mpdf->basepath = $this->depth;
        $mpdf->usepre = false;
        $mpdf->margin_bottom_collapse = true;
        $mpdf->SetDisplayMode('fullpage');

        // $mpdf->setBasePath(empty($this->depth) ? './' : $this->depth);
        $mpdf->SetAutoFont(AUTOFONT_ALL);

        // Temp dir

        $mpdf->ignore_invalid_utf8 = true;
        $mpdf->mirrorMargins = 0;	// don't mirror margins
        $mpdf->WriteHTML($html);
        $mpdf->Output($filename, "F");

        return true;
    }

    function arrangeHtml(&$html, $norendertags = '' ) {

        // add bookmark links
        //$html = preg_replace("/\<a name=(.+?)\>(.+?)\<\/a\>/s",'$2',$html);

        // add bookmark links
        $html = preg_replace_callback("/<h(\d)(.*?)>(.+?)<\/h\\1>/s", array($this, '__pdfHeaderCallback'), $html);
        $html = preg_replace_callback("/<\/div>\s*?<h(" . $this->getConf('PDFHeaderPagebreak') . ")(.*?)>/s", array($this, '__pdfHeaderCallbackPagebreak'), $html);
        $html = preg_replace("/(<img.*?mediacenter.*?\/>)/", "<table style=\"width:100%; border: 0px solid #000;\"><tr><td style=\"text-align: center\">$1</td></tr></table>", $html);
        $html = preg_replace("/<p>(\s*?<table.*?<\/table>\s*?)<\/p>/s", "$1", $html);
        $html = preg_replace_callback("/<pre(.*?)>(.*?)<\/pre>/s", array($this, '__pdfPreCodeCallback'), $html);
        $html = preg_replace_callback("/<a href=\"mailto:(.*?)\".*?>(.*?)<\/a>/s", array($this, '__pdfMailtoCallback'), $html);
		/**/

        $standardReplacer = array (
        // insert a pagebreak for support of WRAP and PAGEBREAK plugins
        							'<br style="page-break-after:always;">' => '<pagebreak />',
                                    '<div class="wrap_pagebreak"></div>' => '<pagebreak />',
                                    '<sup>' => '<sup class="sup">',
                                    '<sub>' => '<sub class="sub">',
                                    '<code>' => '<code class="code">',
        );
        $html = str_replace(array_keys($standardReplacer), array_values($standardReplacer), $html);

        // thanks to Jared Ong
        // Customized to strip all span tags so that the wiki <code> SQL would display properly
        $norender = explode(',',$norendertags);
        $html = $this->strip_only($html, $norender ); //array('span','acronym'));
        $html = $this->strip_htmlencodedchars($html);
        // Customized to strip all span tags so that the wiki <code> SQL would display properly
    }

    function __pdfMailtoCallback($DATA) {
        if ( $DATA[1] == $DATA[2] ) {
            $DATA[2] = $this->deobfuscate($DATA[2]);
        }
        $DATA[1] = $this->deobfuscate($DATA[1]);
        return "<a href=\"mailto:{$DATA[1]}\">{$DATA[2]}</a>";
    }

    function __pdfPreCodeCallback($DATA) {
        
        $code = nl2br($DATA[2]);
        $code = preg_replace_callback("/(^|<br \/>)(\s+)(\S)/s", array($this, '__pdfPreWhitespacesCallback'), $code);
        
        return "\n<pre" . $DATA[1] . ">\n" . $code . "\n</pre>\n";
    }
    
    function __pdfPreWhitespacesCallback( $DATA ) {
        return $DATA[1] . "\n" . str_repeat("&nbsp;", strlen($DATA[2])-($DATA[2]{0}=="\n"?1:0) ) . $DATA[3];
    }
    
    function __pdfHeaderCallback($DATA) {
        $contentText = $this->_xmlEntities(preg_replace("/<\/?.*?>/s", '', $DATA[3])); // Double encoding - has to be decoded in mpdf once more.
        return '<h' . $DATA[1] . $DATA[2] . '>' . $DATA[3] . '<tocentry content="' . $contentText . '" level="' . ($DATA[1]-1) . '" /><bookmark content="' . $contentText . '" level="' . ($DATA[1]-1) . '" /></h' . $DATA[1] . '>';
    }

    function __pdfHeaderCallbackPagebreak($DATA) {
        return '</div>' . "\r\n" . '<pagebreak />' . "\r\n\r\n" . '<h' . $DATA[1] . $DATA[2] . '>';
    }

    // thanks to Jared Ong
    // Custom function for help in stripping span tags
    function strip_only($str, $tags) {
        if(!is_array($tags)) {
            $tags = (strpos($str, '>') !== false ? explode('>', str_replace('<', '', $tags)) : array($tags));
            if(end($tags) == '') array_pop($tags);
        }

        foreach($tags as $tag) $str = preg_replace('#</?'.$tag.'[^>]*>#is', '', $str);
        return $str;
    }
    // Custom function for help in stripping span tags

    // Custom function for help in replacing &#039; &quot; &gt; &lt; &amp;
    function strip_htmlencodedchars($str) {
        $str = str_replace('&#039;', '\'', $str);
        //        $str = str_replace('&quot;', '"', $str);
        //        $str = str_replace('&gt;', '>', $str);
        //        $str = str_replace('&lt;', '<', $str);
        //        $str = str_replace('&amp;', '&', $str);
        return $str;
    }
    // Custom function for help in replacing &#039; &quot; &gt; &lt; &amp;

    /**
     * return an de-obfuscated email address in line with $conf['mailguard'] setting
     */
    function deobfuscate($email) {
        global $conf;

        switch ($conf['mailguard']) {
            case 'visible' :
                $obfuscate = array(' [at] ' => '@', ' [dot] ' => '.', ' [dash] ' => '-');
                return strtr($email, $obfuscate);

            case 'hex' :
                $encode = '';
                $len = strlen($email);
                for ($x=0; $x < $len; $x+=6){
                    $encode .= chr(hexdec($email{$x+3}.$email{($x+4)}));
                }
                return $encode;

            case 'none' :
            default :
                return $email;
        }
    }


    /**
     * print debug info to file if exists
     */
    function _debug($info,$var=null,$level=4){

        if( $this->debugLevel >= $level ) return; // only log certain Debug Levels

        if ( empty( $this->debugFile) ) {
            $this->runtimeException("DebugFile not properly configured. Make sure, it is set, readable and writable. We suggest to use a file in the DokuWiki's media directory.", true);
            $this->debugLevel = 5; // shutdown debug
        }


        $fh = @fopen($this->debugFile, "a+");
        if ( !$fh ) {
            $this->runtimeException("Could not create/open/append logfile: '{$this->debugFile}'", true);
            $this->debugLevel = 5; // shutdown debug
            return;
        }

        switch($level) {
            case 4: $TYPE = "ERROR"; break;
            case 3: $TYPE = " WARN"; break;
            case 2: $TYPE = " INFO"; break;
            case 1: $TYPE = "DEBUG"; break;
            default: $TYPE = " NONE"; break;
        }

        $prepend = "[" . @date('Y-m-d H:i:s') . " $TYPE] ";
        fwrite($fh, $prepend . str_replace("\n", "\n" . $prepend . "\t", trim($info)) . "\n");

        if ( !empty($var) ) {

            if ( is_array($var) ) {
                ob_start();
                print_r($var);
                $content = ob_get_contents();
                ob_end_clean();
            } else {
                $content = $var;
            }
            fwrite($fh, $prepend . "\t" . str_replace("\n", "\n" . $prepend . "\t", str_replace("\r\n", "\n", trim($content))) . "\n");
        }

        fclose($fh);
    }

    function runtimeException($message, $wasDebug=false) {

        if ( empty($message) ) { return; }

        if ( !$this->isAJAX ) {
            print '<b>'.$message.'</b><br />' . "\n";
            if ( $this->firstRE ) {
                print '<b>If this error persists, please contact the server administrator.</b><br />' . "\n";
            }
        } else {
            if ( !$wasDebug ) {
                $this->_debug('Runtime Error: ' . $message, null, 4);
            }

            print 'Runtime Error: ' . $message . "\n";
        }

        $this->firstRE = false;
        return;
    }
}