<?php
if(!defined('DOKU_INC')) define('DOKU_INC',realpath(dirname(__FILE__).'/../../../').'/');
if(!defined('DOKU_PLUGIN')) define('DOKU_PLUGIN',DOKU_INC.'lib/plugins/');
require_once(DOKU_PLUGIN.'admin.php');

/**
 * All DokuWiki plugins to extend the admin function
 * need to inherit from this class
 */
class admin_plugin_siteexport extends DokuWiki_Admin_Plugin {

	var $templateSwitching = false;
    var $pdfExport = false;
    var $usenumberedheading = false;
	
	/**
	 * Constructor
	 */
	function admin_plugin_siteexport(){
		$this->setupLocale();
	}
	
	/**
	 * for backward compatability
	 * @see inc/DokuWiki_Plugin#getInfo()
	 */
    function getInfo(){
        if ( method_exists(parent, 'getInfo')) {
            $info = parent::getInfo();
        }
        return is_array($info) ? $info : confToHash(dirname(__FILE__).'/plugin.info.txt');
    }
	
	/**
	 * return sort order for position in admin menu
	 */
	function getMenuSort() {
		return 100;
	}

	function forAdminOnly(){
		return false;
	}

	/**
	 * handle user request
	 */
	function handle() {

		if ( ! $functions=& plugin_load('preload', 'siteexport') ) return false;
		if ( !$functions->__create_preload_function() ) {
			msg("Can't create preload file. Template switching is not available. Plugin disabling is not available.");
		} else {
			$this->templateSwitching = true;
		}
		
		if ( $functions =& plugin_load('action', 'dw2pdf' ) ) {
		    $this->pdfExport = true;
		}

		if ( $functions =& plugin_load('renderer', 'nodetailsxhtml' ) ) {
		    $this->usenumberedheading = true;
		}
	}

	/**
	 * output appropriate html
	 */
	function html() {
		global $ID, $conf;

		if ( ! $functions=& plugin_load('helper', 'siteexport') ) {
			msg("Can't initialize");
			return false;
		}
		
		$regenerateScript = 'if (!e) var e = window.event; if (e) e.cancelBubble = true; if (e && e.stopPropagation) e.stopPropagation(); var se = new siteexport__start(); se.run(true); return false;';
		$regenerate = array('onchange' => $regenerateScript);
		print $this->plugin_locale_xhtml('intro');

		$form = new Doku_Form('siteexport', null, 'post');
		$form->startFieldset( $this->getLang('startingNamespace') );

		$form->addElement(form_makeTextField('ns', $ID, $this->getLang('ns') . ':', 'ns', null, $regenerate));
		$form->addElement(form_makeTag('br'));
		$form->addElement(form_makeTextField('ens', $ID, $this->getLang('ens') . ':', 'ens', null, $regenerate));
		
		$form->addElement(form_makeTag('br'));
		$changeListener = '$("depthContainer").style.display = this.selectedIndex == 2 ? "block" : "none";';
		$form->addElement(form_makeListboxField('depthType', array( "0" => $this->getLang('depth.pageOnly'), "1" => $this->getLang('depth.allSubNameSpaces'), "2" => $this->getLang('depth.specifiedDepth') ), (empty($_REQUEST['depthType']) ? $this->getLang('depth.allSubNameSpaces') : $_REQUEST['depthType']), $this->getLang('depthType') . ':', 'depthType', null, array_merge(array('class' => 'edit'), array('onchange' => $changeListener . ' ' . $regenerateScript))));
		
		$form->addElement(form_makeTag('br'));
		$form->addElement(form_makeOpenTag("div", array('style' => 'display:' . ($_REQUEST['depthType'] == "3" ? "block" : "none") . ';', 'id' => 'depthContainer')));
		$form->addElement(form_makeTextField('depth', $this->getConf('depth'), $this->getLang('depth') . ':', 'depth', null, $regenerate));
		$form->addElement(form_makeCloseTag("div"));

		$form->addElement(form_makeTag('br'));
		$form->addElement(form_makeOpenTag("div", array('style' => 'display:none;', 'id' => 'depthContainer')));
		$form->addElement(form_makeCheckboxField('exportLinkedPages', 1, $this->getLang('exportLinkedPages') . ':', 'exportLinkedPages', null, array('onchange' => $regenerateScript)));
		$form->addElement(form_makeCloseTag("div"));
				
		$form->endFieldset();
		$form->addElement(form_makeTag('br'));

		$form->startFieldset( $this->getLang('selectYourOptions') );
		$form->addElement(form_makeCheckboxField('absolutePath', 1, $this->getLang('absolutePath') . ':', 'absolutePath', null, $regenerate));
		$form->addElement(form_makeTag('br'));
		$form->addElement(form_makeCheckboxField('exportBody', 1, $this->getLang('exportBody') . ':', 'exportBody', null, $regenerate));
		$form->addElement(form_makeTag('br'));
		$form->addElement(form_makeCheckboxField('addParams', 1, $this->getLang('addParams') . ':', 'addParams', null, array_merge(array('checked' => ($conf['userewrite'] != 1 ? 'checked' : '' ) ), $regenerate)));
		$form->addElement(form_makeTag('br'));
		$form->addElement(form_makeTag('br'));
		$form->addElement(form_makeListboxField('renderer', array_merge(array('','xhtml'), plugin_list('renderer')), '', $this->getLang('renderer') . ':', 'renderer', null, array_merge(array('class' => 'edit'), $regenerate)));

		if ( $this->templateSwitching ) {
			$form->addElement(form_makeTag('br'));
			$form->addElement(form_makeListboxField('template', $functions->__getTemplates(), $conf['template'], $this->getLang('template') . ':', 'template', null, array_merge(array('class' => 'edit'), $regenerate)));
		}

		$form->addElement(form_makeTag('br'));
		$form->addElement(form_makeTag('br'));
		$form->addElement(form_makeCheckboxField('pdfExport', 1, $this->getLang('pdfExport') . ':', 'pdfExport', null, $this->pdfExport ? $regenerate : array_merge(array('disabled' => 'disabled'), $regenerate) ));
		if ( !$this->pdfExport ) {
		    $form->addElement(form_makeOpenTag('p', array('style' => 'color: #a00;' )));
		    $form->addElement('In order to use the PDF export, please ');
		    $form->addElement(form_makeOpenTag('a', array('href' => 'http://www.dokuwiki.org/plugin:dw2pdf', 'alt' => 'install plugin', 'target' => '_blank')));
		    $form->addElement('install the dw2pdf plugin.');
		    $form->addElement(form_makeCloseTag('a'));
		    $form->addElement(form_makeCloseTag('p'));
		}
		
		$form->addElement(form_makeTag('br'));
		$form->addElement(form_makeCheckboxField('usenumberedheading', 1, $this->getLang('usenumberedheading') . ':', 'usenumberedheading', null, $this->usenumberedheading ? $regenerate : array_merge(array('disabled' => 'disabled'), $regenerate) ));
		$form->addElement(form_makeTag('br'));
		if ( !$this->usenumberedheading ) {
		    $form->addElement(form_makeOpenTag('p', array('style' => 'color: #a00;' )));
		    $form->addElement('In order to use numbered headings, please ');
		    $form->addElement(form_makeOpenTag('a', array('href' => 'http://www.dokuwiki.org/plugin:nodetailsxhtml', 'alt' => 'install plugin', 'target' => '_blank')));
		    $form->addElement('install the nodetailsxhtml plugin.');
		    $form->addElement(form_makeCloseTag('a'));
		    $form->addElement(form_makeCloseTag('p'));
		}
		
		$form->endFieldset();
		$form->addElement(form_makeTag('br'));

		$form->startFieldset( $this->getLang('helpCreationOptions') );
		$form->addElement(form_makeCheckboxField('eclipseDocZip', 1, $this->getLang('eclipseDocZip') . ':', 'eclipseDocZip', null, $regenerate));
		$form->addElement(form_makeTag('br'));
		$form->addElement(form_makeCheckboxField('JavaHelpDocZip', 1, $this->getLang('JavaHelpDocZip') . ':', 'JavaHelpDocZip', null, $regenerate));
		$form->addElement(form_makeTag('br'));
		$form->addElement(form_makeCheckboxField('useTocFile', 1, $this->getLang('useTocFile') . ':', 'useTocFile', null, $regenerate));
		$form->addElement(form_makeTag('br'));
		$form->addElement(form_makeCheckboxField('emptyTocElem', 1, $this->getLang('emptyTocElem') . ':', 'emptyTocElem', null, $regenerate));
		$form->addElement(form_makeTag('br'));
		$form->endFieldset();
		$form->addElement(form_makeTag('br'));

		if ( $this->templateSwitching ) {
			$form->startFieldset( $this->getLang('disablePluginsOption') );
			
			$form->addElement(form_makeCheckboxField("disableall", 1, 'Disable All:', "disableall", null, array('onchange' => 'if (!e) var e = window.event; if (e) e.cancelBubble = true; if (e && e.stopPropagation) e.stopPropagation(); var se = new siteexport__start(); se.changePluginState(this); return false;')));
			$form->addElement(form_makeTag('br'));
			$form->addElement(form_makeTag('br'));
				
			list($allPlugins, $enabledPlugins) = $functions->__getPluginList();
			foreach ( $allPlugins as $plugin ) {
				$form->addElement(form_makeCheckboxField("disableplugin[]", $plugin, $plugin . ':', "disableplugin_$plugin", null, array_merge($regenerate, (!in_array($plugin, $enabledPlugins) ? array('checked' => 'checked', 'disabled' => 'disabled') : array() ))));
				$form->addElement(form_makeTag('br'));
			}
			
			$form->endFieldset();
			$form->addElement(form_makeTag('br'));
		}
		
		$form->startFieldset( $this->getLang('startProcess') );

		//$form->addElement(form_makeTextField('copyurl', wl(getNS($ID), array('do' => 'siteexport', 'template' => $conf['template'], 'depthType' => (empty($_REQUEST['depthType']) ? 1 : $_REQUEST['depthType'])), true, '&'), $this->getLang('directDownloadLink') . ':', 'copyurl', null, array('readonly' => 'readonly') ));
		$form->addElement(form_makeTextField('copyurl', "", $this->getLang('directDownloadLink') . ':', 'copyurl', null, array('readonly' => 'readonly') ));
		$form->addElement(form_makeTag('br'));
		$form->addElement(form_makeTextField('wgeturl', "", $this->getLang('wgetURLLink') . ':', 'wgeturl', null, array('readonly' => 'readonly') ));
		$form->addElement(form_makeTag('br'));
		$form->addElement(form_makeButton('submit', 'siteexport', $this->getLang('start') , array('onclick' => 'if (!e) var e = window.event; if (e) e.cancelBubble = true; if (e && e.stopPropagation) e.stopPropagation(); var se = new siteexport__start(); se.run(); return false;', 'style' => 'float:right;') ));
		$form->endFieldset();
		$form->addElement(form_makeTag('br'));

		$form->startFieldset( $this->getLang('status') );
		$form->addElement(form_makeOpenTag('span', array('id' => 'siteexport__out')));

		$form->addElement(form_makeCloseTag('span'));
		$form->addElement(form_makeOpenTag('span'));
		$form->addElement(form_makeTag('img', array('src' => DOKU_BASE.'lib/images/loading.gif', 'id' => 'siteexport__throbber')));
		$form->addElement(form_makeCloseTag('span'));
		$form->endFieldset();
		$form->addElement(form_makeTag('br'));

		$form->printForm();
		
		print <<<OUTPUT
		<script type="text/javascript" charset="utf-8" ><!--//--><![CDATA[//><!--
		addInitEvent(function() {var se = new siteexport__start(); se.run(true);});
		//--><!]]></script>
OUTPUT;
	}
}
//Setup VIM: ex: et ts=4 enc=utf-8 :
